"""Fixtures for aioelectricitymaps tests."""

from collections.abc import AsyncGenerator, Generator

import aiohttp
from aioresponses import aioresponses
import pytest

from aiotankerkoenig import Tankerkoenig


@pytest.fixture(name="tankerkoenig_client")
async def client() -> AsyncGenerator[Tankerkoenig, None]:
    """Return a Spotify client."""
    async with (
        aiohttp.ClientSession() as session,
        Tankerkoenig(
            session=session,
            api_key="abc123",
        ) as tankerkoenig_client,
    ):
        yield tankerkoenig_client


@pytest.fixture(name="responses")
def aioresponses_fixture() -> Generator[aioresponses, None, None]:
    """Return aioresponses fixture."""
    with aioresponses() as mocked_responses:
        yield mocked_responses



"""Tankerkoenig API client."""

from __future__ import annotations

import asyncio
from dataclasses import dataclass
from importlib import metadata
import logging
import socket
from typing import Any, Self

from aiohttp import ClientError, ClientResponseError, ClientSession
from yarl import URL

from .const import GasType, Sort
from .exceptions import (
    TankerkoenigConnectionError,
    TankerkoenigConnectionTimeoutError,
    TankerkoenigError,
    TankerkoenigRateLimitError,
)
from .models import (
    PriceInfo,
    PriceInfoResponse,
    Station,
    StationDetailResponse,
    StationListResponse,
)

_LOGGER = logging.getLogger(__name__)

VERSION = metadata.version(__package__)


def _raise_timeout_error(exception: BaseException) -> None:
    """Raise library specific timeout error."""
    msg = "Timeout occurred while connecting to tankerkoenig.de API"
    raise TankerkoenigConnectionTimeoutError(
        msg,
    ) from exception


@dataclass
class Tankerkoenig:
    """Tankerkoenig API client."""

    api_key: str
    session: ClientSession | None = None

    _close_session: bool = False

    async def _request(self, path: str, params: dict[str, Any]) -> str:
        """Handle request to tankerkoenig.de API."""
        _LOGGER.debug("Requesting %s", path)
        url = URL.build(
            scheme="https",
            host="creativecommons.tankerkoenig.de",
            path=path,
            query={"apikey": self.api_key, **params},
        )

        headers = {
            "Accept": "application/json",
        }

        if self.session is None:
            self.session = ClientSession()
            self._close_session = True
            headers.update(
                {
                    "User-Agent": f"aiotankerkoenig/{VERSION}",
                },
            )

        try:
            async with self.session.get(url, headers=headers) as response:
                content_type = response.headers.get("Content-Type", "")
                text = await response.text()
                _LOGGER.debug(
                    "Response from %s (status: %s, content-type: %s): %s",
                    path,
                    response.status,
                    content_type,
                    text.replace(self.api_key, "**redacted**"),
                )

                response.raise_for_status()
        except TimeoutError as exception:
            _raise_timeout_error(exception)
        except asyncio.CancelledError as exception:
            if "timeout" not in str(exception).lower():
                raise
            _raise_timeout_error(exception)
        except (
            ClientError,
            socket.gaierror,
        ) as exception:
            if isinstance(exception, ClientResponseError) and exception.status == 503:
                raise TankerkoenigRateLimitError from exception
            msg = "Error occurred while communicating with the tankerkoenig.de API"
            raise TankerkoenigConnectionError(msg) from exception

        if "application/json" not in content_type:
            msg = "Unexpected content type from tankerkoenig.de API"
            raise TankerkoenigError(
                msg,
                {"Content-Type": content_type, "response": text},
            )

        return text

    async def nearby_stations(
        self,
        coordinates: tuple[float, float],
        radius: int,
        gas_type: GasType,
        sort: Sort,
    ) -> list[Station]:
        """Get nearby stations."""
        _LOGGER.debug(
            "Fetching nearby stations near (%0.2f, %0.2f) radius=%s type=%s sort=%s",
            coordinates[0],
            coordinates[1],
            radius,
            gas_type,
            sort,
        )
        result = await self._request(
            path="/json/list.php",
            params={
                "lat": coordinates[0],
                "lng": coordinates[1],
                "rad": radius,
                "type": gas_type,
                "sort": sort,
            },
        )
        return StationListResponse.from_json(result).stations

    async def station_details(
        self,
        station_id: str,
    ) -> Station:
        """Get station details."""
        _LOGGER.debug("Fetching station details for %s", station_id)
        result = await self._request(
            path="/json/detail.php",
            params={
                "id": station_id,
            },
        )
        return StationDetailResponse.from_json(result).station

    async def prices(
        self,
        station_ids: list[str],
    ) -> dict[str, PriceInfo]:
        """Get station details."""
        _LOGGER.debug("Fetching prices for %d station(s)", len(station_ids))
        result = await self._request(
            path="/json/prices.php",
            params={
                "ids": ",".join(station_ids),
            },
        )
        return PriceInfoResponse.from_json(result).prices

    async def close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()

    async def __aenter__(self) -> Self:
        """Async enter."""
        return self

    async def __aexit__(self, *_exc_info: object) -> None:
        """Async exit.    _exc_info: Exec type."""
        await self.close()



"""Test the tankerkoenig API client."""

import asyncio

import aiohttp
from aioresponses import aioresponses
import pytest
from syrupy import SnapshotAssertion

from aiotankerkoenig import (
    GasType,
    Sort,
    Tankerkoenig,
    TankerkoenigConnectionError,
    TankerkoenigConnectionTimeoutError,
    TankerkoenigError,
    TankerkoenigInvalidKeyError,
    TankerkoenigRateLimitError,
)
from tests import load_fixture

TANKERKOENIG_ENDPOINT = "https://creativecommons.tankerkoenig.de"


async def test_unexpected_server_response(
    responses: aioresponses,
    tankerkoenig_client: Tankerkoenig,
) -> None:
    """Test handling unexpected response."""
    responses.get(
        f"{TANKERKOENIG_ENDPOINT}/json/detail.php?apikey=abc123&id=1",
        status=200,
        headers={"Content-Type": "plain/text"},
        body="Yes",
    )
    with pytest.raises(TankerkoenigError):
        await tankerkoenig_client.station_details(station_id="1")


async def test_non_200_response(
    responses: aioresponses,
    tankerkoenig_client: Tankerkoenig,
) -> None:
    """Test handling unexpected response."""
    responses.get(
        f"{TANKERKOENIG_ENDPOINT}/json/detail.php?apikey=abc123&id=1",
        status=500,
        headers={"Content-Type": "plain/text"},
        body="Yes",
    )
    with pytest.raises(TankerkoenigConnectionError):
        await tankerkoenig_client.station_details(station_id="1")


async def test_api_rate_limit(
    responses: aioresponses,
    tankerkoenig_client: Tankerkoenig,
) -> None:
    """Test handling api rate limit response."""
    responses.get(
        f"{TANKERKOENIG_ENDPOINT}/json/detail.php?apikey=abc123&id=1",
        status=503,
        headers={"Content-Type": "plain/text"},
        body="Yes",
    )
    with pytest.raises(TankerkoenigRateLimitError):
        await tankerkoenig_client.station_details(station_id="1")


@pytest.mark.parametrize(
    "message",
    [
        "API-Key existiert nicht",
        "APIKey existiert nicht.",
    ],
)
async def test_invalid_apikey_response(
    responses: aioresponses,
    tankerkoenig_client: Tankerkoenig,
    message: str,
) -> None:
    """Test handling of invalid api key response."""
    responses.get(
        f"{TANKERKOENIG_ENDPOINT}/json/detail.php?apikey=abc123&id=1",
        status=200,
        body='{"status": "error", "ok": false, "message": "' + message + '"}',
    )
    with pytest.raises(TankerkoenigInvalidKeyError):
        await tankerkoenig_client.station_details(station_id="1")


async def test_general_error_response(
    responses: aioresponses,
    tankerkoenig_client: Tankerkoenig,
) -> None:
    """Test handling of a general error response."""
    responses.get(
        f"{TANKERKOENIG_ENDPOINT}/json/detail.php?apikey=abc123&id=1",
        status=200,
        body='{"status": "error", "ok": false, "message": "Idk what happened!"}',
    )
    with pytest.raises(TankerkoenigError):
        await tankerkoenig_client.station_details(station_id="1")


async def test_putting_in_own_session(
    responses: aioresponses,
) -> None:
    """Test putting in own session."""
    responses.get(
        f"{TANKERKOENIG_ENDPOINT}/json/detail.php?apikey=abc123&id=1",
        status=200,
        body=load_fixture("detail.json"),
    )
    async with aiohttp.ClientSession() as session:
        tankerkoenig = Tankerkoenig(api_key="abc123", session=session)
        await tankerkoenig.station_details(station_id="1")
        assert tankerkoenig.session is not None
        assert not tankerkoenig.session.closed
        await tankerkoenig.close()
        assert not tankerkoenig.session.closed


async def test_creating_own_session(
    responses: aioresponses,
) -> None:
    """Test creating own session."""
    responses.get(
        f"{TANKERKOENIG_ENDPOINT}/json/detail.php?apikey=abc123&id=1",
        status=200,
        body=load_fixture("detail.json"),
    )
    tankerkoenig = Tankerkoenig(api_key="abc123")
    await tankerkoenig.station_details(station_id="1")
    assert tankerkoenig.session is not None
    assert not tankerkoenig.session.closed
    await tankerkoenig.close()
    assert tankerkoenig.session.closed


async def test_station_details(
    responses: aioresponses,
    tankerkoenig_client: Tankerkoenig,
    snapshot: SnapshotAssertion,
) -> None:
    """Test retrieving station details."""
    responses.get(
        f"{TANKERKOENIG_ENDPOINT}/json/detail.php?apikey=abc123&id=1",
        status=200,
        body=load_fixture("detail.json"),
    )
    assert await tankerkoenig_client.station_details(station_id="1") == snapshot


async def test_nearby_stations(
    responses: aioresponses,
    tankerkoenig_client: Tankerkoenig,
    snapshot: SnapshotAssertion,
) -> None:
    """Test retrieving nearby stations."""
    responses.get(
        f"{TANKERKOENIG_ENDPOINT}/json/list.php?apikey=abc123&lat=53.1&lng=7.8&rad=5&sort=dist&type=all",
        status=200,
        body=load_fixture("list.json"),
    )
    assert (
        await tankerkoenig_client.nearby_stations(
            coordinates=(53.1, 7.8),
            radius=5,
            gas_type=GasType.ALL,
            sort=Sort.DISTANCE,
        )
        == snapshot
    )


async def test_prices(
    responses: aioresponses,
    tankerkoenig_client: Tankerkoenig,
    snapshot: SnapshotAssertion,
) -> None:
    """Test retrieving nearby stations."""
    responses.get(
        f"{TANKERKOENIG_ENDPOINT}/json/prices.php?apikey=abc123&ids=4fd02fd5-0a4c-46b3-a9b9-a488a62454be",
        status=200,
        body=load_fixture("prices.json"),
    )
    assert (
        await tankerkoenig_client.prices(
            station_ids=["4fd02fd5-0a4c-46b3-a9b9-a488a62454be"],
        )
        == snapshot
    )


async def test_timeout(
    responses: aioresponses,
) -> None:
    """Test request timeout."""
    responses.get(
        f"{TANKERKOENIG_ENDPOINT}/json/detail.php?apikey=abc123&id=1",
        timeout=True,
    )
    async with Tankerkoenig(
        api_key="abc123",
    ) as tankerkoenig_client:
        with pytest.raises(TankerkoenigConnectionTimeoutError):
            await tankerkoenig_client.station_details(station_id="1")


async def test_timeout_cancellation() -> None:
    """Test request timeout raised as a cancelled task."""

    class TimeoutSession:
        """Session raising a timeout cancellation from get."""

        def get(
            self,
            *_args: object,
            **_kwargs: object,
        ) -> "TimeoutSession":
            """Return self as an async context manager."""
            return self

        async def __aenter__(self) -> None:
            """Raise cancellation with a timeout reason."""
            msg = "Global task timeout"
            raise asyncio.CancelledError(msg)

        async def __aexit__(self, *_exc_info: object) -> None:
            """Async exit."""

    tankerkoenig_client = Tankerkoenig(
        api_key="abc123",
        session=TimeoutSession(),  # type: ignore[arg-type]
    )
    with pytest.raises(TankerkoenigConnectionTimeoutError):
        await tankerkoenig_client.station_details(station_id="1")


async def test_non_timeout_cancellation() -> None:
    """Test request cancellation without timeout reason."""

    class CancelledSession:
        """Session raising a non-timeout cancellation from get."""

        def get(
            self,
            *_args: object,
            **_kwargs: object,
        ) -> "CancelledSession":
            """Return self as an async context manager."""
            return self

        async def __aenter__(self) -> None:
            """Raise cancellation without a timeout reason."""
            raise asyncio.CancelledError

        async def __aexit__(self, *_exc_info: object) -> None:
            """Async exit."""

    tankerkoenig_client = Tankerkoenig(
        api_key="abc123",
        session=CancelledSession(),  # type: ignore[arg-type]
    )
    with pytest.raises(asyncio.CancelledError):
        await tankerkoenig_client.station_details(station_id="1")



"""Exceptions for the aiotankerkoenig client."""

from __future__ import annotations


class TankerkoenigError(Exception):
    """Base exception class for all exceptions raised by this library."""


class TankerkoenigConnectionError(TankerkoenigError):
    """Raised when a connection error occurs."""


class TankerkoenigConnectionTimeoutError(TankerkoenigConnectionError):
    """Raised when a connection times out."""


class TankerkoenigInvalidKeyError(TankerkoenigError):
    """Raised when the API key is invalid."""


class TankerkoenigRateLimitError(TankerkoenigError):
    """Raised when the API rate limit is reached."""



"""Tankerkoenig API client."""

from __future__ import annotations

from .aiotankerkoenig import Tankerkoenig
from .const import GasType, Sort, Status
from .exceptions import (
    TankerkoenigConnectionError,
    TankerkoenigConnectionTimeoutError,
    TankerkoenigError,
    TankerkoenigInvalidKeyError,
    TankerkoenigRateLimitError,
)
from .models import PriceInfo, Station

__all__ = [
    "GasType",
    "PriceInfo",
    "Sort",
    "Station",
    "Status",
    "Tankerkoenig",
    "TankerkoenigConnectionError",
    "TankerkoenigConnectionTimeoutError",
    "TankerkoenigError",
    "TankerkoenigInvalidKeyError",
    "TankerkoenigRateLimitError",
]



"""Models for the aiotankerkoenig client."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Self

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin

from .const import Status
from .exceptions import TankerkoenigError, TankerkoenigInvalidKeyError


@dataclass(frozen=True, slots=True, kw_only=True)
class TankerkoenigResponse(DataClassORJSONMixin):
    """Base class for all responses."""

    @classmethod
    def __pre_deserialize__(
        cls: type[Self],
        d: dict[Any, Any],
    ) -> dict[Any, Any]:
        """Raise when response was unexpected."""
        if d.get("ok"):
            return d

        message = d.get("message", "")
        if any(x in message.lower() for x in ("api-key", "apikey")):
            msg = "tankerkoenig.de API responded with an invalid key error"
            raise TankerkoenigInvalidKeyError(
                msg,
                {"response": d},
            )

        msg = f"tankerkoenig.de API responded with an error: {message}"
        raise TankerkoenigError(
            msg,
            {"response": d.pop("message")},
        )


@dataclass(frozen=True, slots=True, kw_only=True)
class StationListResponse(TankerkoenigResponse):
    """Class representing a station list response."""

    stations: list[Station]


@dataclass(frozen=True, slots=True, kw_only=True)
class StationDetailResponse(TankerkoenigResponse):
    """Class representing a station response."""

    station: Station


@dataclass(frozen=True, slots=True, kw_only=True)
class PriceInfoResponse(TankerkoenigResponse):
    """Class representing a station price info response."""

    prices: dict[str, PriceInfo]


@dataclass(frozen=True, slots=True, kw_only=True)
class OpeningTime:
    """Class representing a station opening time."""

    end: str
    start: str
    text: str


@dataclass(frozen=True, slots=True, kw_only=True)
class Station:
    """Class representing a station."""

    id: str
    name: str
    house_number: str = field(metadata=field_options(alias="houseNumber"))
    place: str
    post_code: int = field(metadata=field_options(alias="postCode"))
    street: str
    lat: float
    lng: float

    # only available in detail view
    brand: str | None = None
    diesel: float | None = None
    e10: float | None = None
    e5: float | None = None
    is_open: bool | None = field(metadata=field_options(alias="isOpen"), default=None)
    overrides: Any | None = None
    whole_day: bool | None = field(
        metadata=field_options(alias="wholeDay"),
        default=None,
    )
    state: str | None = None
    opening_times: list[OpeningTime] | None = field(
        metadata=field_options(alias="openingTimes"),
        default=None,
    )
    distance: float | None = field(metadata=field_options(alias="dist"), default=None)


@dataclass(frozen=True, slots=True, kw_only=True)
class PriceInfo:
    """Class representing a station price info."""

    status: Status = Status.UNKNOWN
    e5: float | None = None
    e10: float | None = None
    diesel: float | None = None



"""Constants for the aiotankerkoenig."""

from enum import StrEnum


class GasType(StrEnum):
    """Gas type."""

    ALL = "all"
    DIESEL = "diesel"
    E5 = "e5"
    E10 = "e10"


class Sort(StrEnum):
    """Sort type."""

    DISTANCE = "dist"
    PRICE = "price"
    TIME = "time"


class Status(StrEnum):
    """Status type."""

    OPEN = "open"
    CLOSED = "closed"
    UNKNOWN = "unknown"
